package pp.s1184725.boppi;

import java.nio.file.Files;
import java.util.*;
import java.util.logging.Logger;

import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.tree.*;
import org.apache.commons.lang3.StringUtils;

import pp.iloc.Assembler;
import pp.iloc.eval.Machine;
import pp.iloc.model.*;
import pp.s1184725.boppi.antlr.*;
import pp.s1184725.boppi.antlr.BoppiParser.*;
import pp.s1184725.boppi.type.*;
import pp.s1184725.boppi.util.RegisterPool;

/**
 * The generator stage of the Boppi toolchain.
 * 
 * @author Frank Wibbelink
 */
public class BoppiGenerator extends BoppiBaseVisitor<Reg> {
	private static final int ARBASESIZE = 16, FUNCREF_SIZE = 12;
	private static final Num ZERO = new Num(0);
	private static final Reg ZERO_REG = new Reg("zero");
	private static final Num OFFSET_ARP = new Num(-4), OFFSET_RETURN_ADDR = new Num(-8),
			OFFSET_RETURN_VAL = new Num(-12), OFFSET_AL = new Num(-16), OFFSET_FUNCREF_ADDR = ZERO,
			OFFSET_FUNCREF_ARP = new Num(4), OFFSET_FUNCREF_ARSIZE = new Num(8);

	private Program prog;
	private Annotations an;
	private final Reg tempArp = new Reg("ART"), arp = Machine.ARP_REG;
	private Logger logger;
	private RegisterPool regPool;
	private int labels = 0;

	/**
	 * Maps operator tokens to ILOC opcodes.
	 */
	private static Map<Integer, OpCode> ops = new HashMap<Integer, OpCode>() {
		private static final long serialVersionUID = 8979722313842633807L;

		{
			put(BoppiLexer.AND, OpCode.and);
			put(BoppiLexer.DIVIDE, OpCode.div);
			put(BoppiLexer.EQ, OpCode.cmp_EQ);
			put(BoppiLexer.GT, OpCode.cmp_GT);
			put(BoppiLexer.GTE, OpCode.cmp_GE);
			put(BoppiLexer.LEQ, OpCode.cmp_LE);
			put(BoppiLexer.LT, OpCode.cmp_LT);
			put(BoppiLexer.MINUS, OpCode.sub);
			put(BoppiLexer.MULTIPLY, OpCode.mult);
			put(BoppiLexer.NEQ, OpCode.cmp_NE);
			put(BoppiLexer.PLUS, OpCode.add);
			put(BoppiLexer.OR, OpCode.or);
		}
	};

	/**
	 * Creates a new generator with the given variable annotations and attaches
	 * the given logger.
	 * 
	 * @param ast
	 *            the program's AST
	 * @param annotations
	 *            the annotation object to use during generation
	 * @param logger
	 *            the logger to report errors to
	 * @return an ILOC program generated from the AST
	 */
	public static Program generateProgram(ParseTree ast, Annotations annotations, Logger logger) {
		BoppiGenerator generator = new BoppiGenerator(annotations, logger);
		generator.visit(ast);
		return generator.prog;
	}

	protected BoppiGenerator(Annotations annotations, Logger logger) {
		try {
			String memlib = new String(Files.readAllBytes(ToolChain.PATH.resolve("memlib.iloc")));
			prog = Assembler.instance().assemble(memlib);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		this.an = annotations;
		this.regPool = new RegisterPool(logger);
		this.logger = logger;
	}

	private Reg getVar(VariableContext node) {
		Variable<Type> var = an.variables.get(node);

		int functionDepth = an.function.get(node).getFunctionDepth();

		if (var.getDepth() < functionDepth) {
			emit("travelling ALs", OpCode.i2i, arp, tempArp);
			for (int i = var.getDepth(); i < functionDepth; i++)
				emit(StringUtils.repeat(' ', i) + "\\", OpCode.loadAI, tempArp, OFFSET_AL, tempArp);

			return tempArp;
		} else
			return arp;
	}

	/**
	 * Constructs an operation from the parameters and adds it to the program
	 * under construction.
	 */
	private Op emit(String comment, Label label, OpCode opCode, Operand... args) {
		Op result = new Op(label, opCode, args);
		result.setComment(String.format("%s", comment));
		this.prog.addInstr(result);
		return result;
	}

	/**
	 * Constructs an operation from the parameters and adds it to the program
	 * under construction.
	 */
	private Op emit(String comment, OpCode opCode, Operand... args) {
		return emit(comment, (Label) null, opCode, args);
	}

	/**
	 * Increments the chain of ARs.
	 * 
	 * @param temp
	 *            temporary register to use
	 * @param ar
	 *            register containing the top-level AR
	 */
	private void incARReferences(Reg temp, Reg ar) {
		String comment = "AR incRef";
		Label loop = makeLabel("aril"), done = makeLabel("arid");
		emit(comment, OpCode.i2i, ar, tempArp);
		emit(comment, OpCode.cmp_NE, tempArp, ZERO_REG, temp);
		emit(comment, OpCode.cbr, temp, loop, done);
		emit(comment, loop, OpCode.loadI, new Num(prog.size() + 5), temp);
		emit(comment, OpCode.push, temp);
		emit(comment, OpCode.subI, tempArp, new Num(ARBASESIZE), temp);
		emit(comment, OpCode.push, temp);
		emit(comment, OpCode.jumpI, new Label("memaddref"));

		emit(comment, OpCode.loadAI, tempArp, OFFSET_AL, tempArp);
		emit(comment, OpCode.cmp_NE, tempArp, ZERO_REG, temp);
		emit(comment, OpCode.cbr, temp, loop, done);
		emit(comment, done, OpCode.nop);
	}

	/**
	 * Decrements the chain of ARs.
	 * 
	 * @param temp
	 *            temporary register to use
	 * @param ar
	 *            register containing the top-level AR
	 */
	private void decARReferences(Reg temp, Reg ar) {
		String comment = "AR decRef";
		Label loop = makeLabel("ardl"), done = makeLabel("ardd");
		emit(comment, OpCode.i2i, ar, tempArp);
		emit(comment, OpCode.cmp_NE, tempArp, ZERO_REG, temp);
		emit(comment, OpCode.cbr, temp, loop, done);
		emit(comment, loop, OpCode.loadI, new Num(prog.size() + 5), temp);
		emit(comment, OpCode.push, temp);
		emit(comment, OpCode.subI, tempArp, new Num(ARBASESIZE), temp);
		emit(comment, OpCode.push, temp);
		emit(comment, OpCode.jumpI, new Label("memfree"));

		emit(comment, OpCode.loadAI, tempArp, OFFSET_AL, tempArp);
		emit(comment, OpCode.cmp_NE, tempArp, ZERO_REG, temp);
		emit(comment, OpCode.cbr, temp, loop, done);
		emit(comment, done, OpCode.nop);
	}

	/**
	 * Allocates a chunk of memory and returns a pointer to the memory in the
	 * given register.
	 * 
	 * @param temp
	 *            temporary register to use
	 * @param size
	 *            the size of the memory chunk
	 */
	private void malloc(Reg temp, int size) {
		emit("malloc", OpCode.loadI, new Num(prog.size() + 5), temp);
		emit("malloc", OpCode.push, temp);
		emit("malloc", OpCode.loadI, new Num(size), temp);
		emit("malloc", OpCode.push, temp);
		emit("malloc", OpCode.jumpI, new Label("memalloc"));
		emit("malloc", OpCode.pop, temp);
	}

	/**
	 * Allocates a chunk of memory and returns a pointer to the memory in the
	 * given register.
	 * 
	 * @param temp
	 *            temporary register to use
	 * @param size
	 *            the register containing the size of the memory chunk
	 */
	private void malloc(Reg temp, Reg size) {
		emit("malloc", OpCode.loadI, new Num(prog.size() + 4), temp);
		emit("malloc", OpCode.push, temp);
		emit("malloc", OpCode.push, size);
		emit("malloc", OpCode.jumpI, new Label("memalloc"));
		emit("malloc", OpCode.pop, temp);
	}

	/**
	 * Frees a chunk of memory pointed at by {@code ptr}.
	 * 
	 * @param temp
	 *            temporary register to use
	 * @param ptr
	 *            the register containing the memory pointer
	 */
	private void free(Reg temp, Reg ptr) {
		emit("free", OpCode.loadI, new Num(prog.size() + 4), temp);
		emit("free", OpCode.push, temp);
		emit("free", OpCode.push, ptr);
		emit("free", OpCode.jumpI, new Label("memfree"));
	}

	/**
	 * Increments the reference count of a memory chunk.
	 * 
	 * @param temp
	 *            temporary register to use
	 * @param ptr
	 *            the register containing the memory pointer
	 */
	private void incRef(Reg temp, Reg ptr) {
		emit("memaddref", OpCode.loadI, new Num(prog.size() + 4), temp);
		emit("memaddref", OpCode.push, temp);
		emit("memaddref", OpCode.push, ptr);
		emit("memaddref", OpCode.jumpI, new Label("memaddref"));
	}

	/**
	 * Returns an error message for a given parse tree node.
	 * 
	 * @param ctx
	 *            the parse tree node at which the error occurred
	 * @param message
	 *            the error message
	 * @param args
	 *            arguments for the message, see {@link String#format}
	 */
	private String getError(ParserRuleContext node, String message, Object... args) {
		int line = node.getStart().getLine();
		int column = node.getStart().getCharPositionInLine();
		return String.format("Line %d:%d - %s", line, column, String.format(message, args));
	}

	private Label makeLabel(String prefix) {
		return new Label(prefix + (labels++));
	}

	private void incrementReference(Type type, Reg addr) {
		regPool.withReg((temp) -> {
			String comment = "add new reference";

			incRef(temp, addr);

			if (type instanceof FunctionType) {
				emit(comment, OpCode.loadAI, addr, OFFSET_FUNCREF_ARP, addr);

				incARReferences(temp, addr);
			}
		});
	}

	private void decrementReference(Type type, Reg addr) {
		regPool.withReg((temp) -> {
			Label isNull = makeLabel("ynul"), notNull = makeLabel("nnul");
			String comment = "remove old reference";

			emit(comment, OpCode.cmp_EQ, addr, ZERO_REG, temp);
			emit(comment, OpCode.cbr, temp, isNull, notNull);

			emit(comment, notNull, OpCode.nop);

			free(temp, addr);

			if (type instanceof FunctionType) {
				emit(comment, OpCode.loadAI, addr, OFFSET_FUNCREF_ARP, addr);

				decARReferences(temp, addr);
			}

			emit(comment, isNull, OpCode.nop);
		});
	}

	private void cleanScope(FunctionScope<Type> scope) {
		regPool.withReg((addr, ft) -> {
			for (Variable<Type> var : scope.getVars()) {
				Type type = var.getType();

				if (type instanceof ReferenceType) {
					Label isNull = makeLabel("ynul"), notNull = makeLabel("nnul");
					String comment = "remove reference";
					emit(comment + " get var", OpCode.loadAI, arp, new Num(var.getOffset()), addr);

					emit(comment, OpCode.cmp_EQ, addr, ZERO_REG, ft);
					emit(comment, OpCode.cbr, ft, isNull, notNull);

					emit(comment, notNull, OpCode.nop);

					free(ft, addr);

					if (type instanceof FunctionType) {
						emit(comment, OpCode.loadAI, addr, OFFSET_FUNCREF_ARP, addr);

						decARReferences(ft, addr);
					}

					emit(comment, isNull, OpCode.nop);
				}
			}

		});
	}

	@Override
	public Reg visit(ParseTree tree) {
		try {
			return super.visit(tree);
		} catch (Exception e) {
			e.printStackTrace();
			System.out.println(prog.prettyPrint());
			if (tree instanceof ParserRuleContext)
				logger.severe(getError((ParserRuleContext) tree, "%s: %s", e.getClass().getName(), e.getMessage()));
			else
				logger.severe(String.format("Error %s: %s", e.getClass().getName(), e.getMessage()));
			return null;
		}
	}

	@Override
	public Reg visitAssign(AssignContext ctx) {

		Reg result = visit(ctx.singleExpr());

		regPool.blockReg(result, () -> {
			Type type = an.types.get(ctx);

			if (type instanceof ReferenceType) {
				regPool.withReg((addr) -> {
					emit("remove old ref for " + ctx.variable().getText(), OpCode.loadAI, getVar(ctx.variable()),
							new Num(an.variables.get(ctx.variable()).getOffset()), addr);
					decrementReference(an.types.get(ctx), addr);
				});
			}

			if (SimpleType.CHAR.equals(type))
				emit("to " + ctx.variable().getText(), OpCode.cstoreAI, result, getVar(ctx.variable()),
						new Num(an.variables.get(ctx.variable()).getOffset()));
			else
				emit("to " + ctx.variable().getText(), OpCode.storeAI, result, getVar(ctx.variable()),
						new Num(an.variables.get(ctx.variable()).getOffset()));

			if (type instanceof ReferenceType) {
				regPool.withReg((addr) -> {
					emit("add new ref for " + ctx.variable().getText(), OpCode.loadAI, getVar(ctx.variable()),
							new Num(an.variables.get(ctx.variable()).getOffset()), addr);

					incrementReference(an.types.get(ctx), addr);
				});
			}
		});

		return result;
	}

	@Override
	public Reg visitBlock(BlockContext ctx) {
		return visit(ctx.expr());
	}

	@Override
	public Reg visitCall(CallContext ctx) {

		String base = "call " + ctx.variable().getText() + " - ";
		Variable<Type> function = an.variables.get(ctx.variable());
		TupleType parameters = (TupleType) ((FunctionType) function.getType()).getParameter();

		visit(ctx.variable());

		Reg result = regPool.withReg((ar) -> {
			regPool.withReg((rt) -> {
				emit(base + "load function reference", OpCode.loadAI, getVar(ctx.variable()),
						new Num(an.variables.get(ctx.variable()).getOffset()), rt);
				emit(base + "load AR size", OpCode.loadAI, rt, OFFSET_FUNCREF_ARSIZE, rt);
				malloc(ar, rt);
			});
			emit(base + "shift AR", OpCode.addI, ar, new Num(ARBASESIZE), ar);

			for (int i = 0; i < ctx.expr().size(); i++) {
				Reg exprReg = visit(ctx.expr(i));

				if (SimpleType.CHAR.equals(parameters.get(i)))
					emit(base + "store param " + i, OpCode.cstoreAI, exprReg, ar, new Num(parameters.getOffset(i)));
				else
					emit(base + "store param " + i, OpCode.storeAI, exprReg, ar, new Num(parameters.getOffset(i)));
			}

			Stack<Reg> inUse = new Stack<>();
			for (Reg reg : regPool.getInUse()) {
				if (!reg.equals(ar)) {
					inUse.push(reg);
					emit(base + "register save " + reg.getName(), OpCode.push, reg);
				}
			}

			Reg res = regPool.withReg((r1, r2) -> {
				emit(base + "load function reference", OpCode.loadAI, getVar(ctx.variable()),
						new Num(an.variables.get(ctx.variable()).getOffset()), r2);
				emit(base + "link caller ARP", OpCode.storeAI, arp, ar, OFFSET_ARP);
				emit(base + "load AL", OpCode.loadAI, r2, OFFSET_FUNCREF_ARP, r1);
				emit(base + "link AL", OpCode.storeAI, r1, ar, OFFSET_AL);

				emit("add ref for callee's AL", OpCode.loadAI, ar, OFFSET_AL, tempArp);
				incARReferences(r1, tempArp);

				emit(base + "load return address", OpCode.loadI, new Num(prog.size() + 5), r1);
				emit(base + "set return address", OpCode.storeAI, r1, ar, OFFSET_RETURN_ADDR);
				emit(base + "move ARP", OpCode.i2i, ar, arp);

				emit(base + "load target address", OpCode.loadAI, r2, OFFSET_FUNCREF_ADDR, r1);
				emit(base + "execute", OpCode.jump, r1);

				emit(base + "load result", OpCode.loadAI, arp, OFFSET_RETURN_VAL, r2);
			});

			while (!inUse.isEmpty()) {
				Reg reg = inUse.pop();
				emit(base + "register unsave " + reg.getName(), OpCode.pop, reg);
			}

			return res;

		});

		regPool.withReg(result, (temp) -> {
			decARReferences(temp, arp);
		});

		emit(base + "reset ARP", OpCode.loadAI, arp, OFFSET_ARP, arp);

		return result;
	}

	/**
	 * Declare named function:
	 * <ol>
	 * <li>generate jump over function body</li>
	 * <li>generate entry point</li>
	 * <li>visit function body</li>
	 * <li>generate return</li>
	 * <li>generate allocate function reference</li>
	 * <li>generate store entry point and AR to function reference</li>
	 * <li>generate store function reference in variable</li>
	 * </ol>
	 */
	@Override
	public Reg visitDeclareFunction(DeclareFunctionContext ctx) {

		String base = "define " + ctx.IDENTIFIER(0).getText() + " - ";
		Label skip = makeLabel("s");
		emit(base + "jump over body", OpCode.jumpI, skip);
		int targetAddress = emit(base + "entry point", OpCode.nop).getLine();

		Reg result = visit(ctx.body);
		emit(base + "move result", OpCode.storeAI, result, arp, OFFSET_RETURN_VAL);

		cleanScope(an.function.get(ctx));

		regPool.withReg((r1, r2) -> {
			emit(base + "load return address", OpCode.loadAI, arp, OFFSET_RETURN_ADDR, r1);
			emit(base + "go to return address", OpCode.jump, r1);

			emit(base + "skip target", skip, OpCode.nop);
			malloc(r2, FUNCREF_SIZE);
			emit(base + "load target address", OpCode.loadI, new Num(targetAddress), r1);
			emit(base + "set target address", OpCode.storeAI, r1, r2, OFFSET_FUNCREF_ADDR);
			emit(base + "copy ARP", OpCode.storeAI, arp, r2, OFFSET_FUNCREF_ARP);
			emit(base + "load AR size", OpCode.loadI, new Num(ARBASESIZE + an.function.get(ctx).getLocalDataSize()),
					r1);
			emit(base + "set AR size", OpCode.storeAI, r1, r2, OFFSET_FUNCREF_ARSIZE);

			emit(base + "set function reference", OpCode.storeAI, r2, arp, new Num(an.variables.get(ctx).getOffset()));
			incARReferences(r2, arp);
		});

		return null;
	}

	@Override
	public Reg visitExpr(ExprContext ctx) {

		for (int i = 0; i < ctx.singleExpr().size() - 1; i++) {
			ParserRuleContext expr = ctx.singleExpr(i);
			Reg result = visit(expr);

			if (result != null)
				regPool.blockReg(result, () -> {
					Type type = an.types.get(expr);

					if (type instanceof ReferenceType)
						decrementReference(type, result);
				});
		}
		ParserRuleContext last = ctx.singleExpr(ctx.singleExpr().size() - 1);

		return visit(last);
	}

	@Override
	public Reg visitIf(IfContext ctx) {
		Label toTrue = makeLabel("if_t");
		Label toFalse = makeLabel("if_f");
		Label toEnd = makeLabel("if_e");
		Reg cond = visit(ctx.cond);

		if (an.types.get(ctx) != SimpleType.VOID) {
			return regPool.withReg((target) -> {
				emit("", OpCode.cbr, cond, toTrue, toFalse);

				emit("", toTrue, OpCode.nop);
				emit("result", OpCode.i2i, visit(ctx.onTrue), target);
				emit("", OpCode.jumpI, toEnd);

				emit("", toFalse, OpCode.nop);
				emit("result", OpCode.i2i, visit(ctx.onFalse), target);
				emit("end target", toEnd, OpCode.nop);
			});

		}

		if (ctx.onFalse == null) {
			emit("", OpCode.cbr, cond, toTrue, toEnd);
			emit("", toTrue, OpCode.nop);
			visit(ctx.onTrue);
			emit("end target", toEnd, OpCode.nop);
		} else {
			emit("", OpCode.cbr, cond, toTrue, toFalse);
			emit("", toTrue, OpCode.nop);
			visit(ctx.onTrue);
			emit("", OpCode.jumpI, toEnd);
			emit("", toFalse, OpCode.nop);
			visit(ctx.onFalse);
			emit("end target", toEnd, OpCode.nop);
		}

		return null;
	}

	@Override
	public Reg visitInfix1(Infix1Context ctx) {
		Reg lhs = visit(ctx.lhs);
		Reg rhs = regPool.blockReg(lhs, () -> visit(ctx.rhs));
		emit(ctx.getChild(1).getText(), ops.get(ctx.op.getType()), lhs, rhs, lhs);

		return lhs;
	}

	@Override
	public Reg visitInfix2(Infix2Context ctx) {
		Reg lhs = visit(ctx.lhs);
		Reg rhs = regPool.blockReg(lhs, () -> visit(ctx.rhs));
		emit(ctx.getChild(1).getText(), ops.get(ctx.op.getType()), lhs, rhs, lhs);

		return lhs;
	}

	@Override
	public Reg visitInfix3(Infix3Context ctx) {
		Reg lhs = visit(ctx.lhs);
		Reg rhs = regPool.blockReg(lhs, () -> visit(ctx.rhs));
		emit(ctx.getChild(1).getText(), ops.get(ctx.op.getType()), lhs, rhs, lhs);

		return lhs;
	}

	@Override
	public Reg visitInfix4(Infix4Context ctx) {
		Reg lhs = visit(ctx.lhs);
		Reg rhs = regPool.blockReg(lhs, () -> visit(ctx.rhs));
		emit(ctx.getChild(1).getText(), ops.get(ctx.AND().getSymbol().getType()), lhs, rhs, lhs);

		return lhs;
	}

	@Override
	public Reg visitInfix5(Infix5Context ctx) {
		Reg lhs = visit(ctx.lhs);
		Reg rhs = regPool.blockReg(lhs, () -> visit(ctx.rhs));
		emit(ctx.getChild(1).getText(), ops.get(ctx.OR().getSymbol().getType()), lhs, rhs, lhs);

		return lhs;
	}

	@Override
	public Reg visitLiteralBoolean(LiteralBooleanContext ctx) {
		return regPool.withReg((reg) -> {
			emit(ctx.getText(), OpCode.loadI, new Num(ctx.TRUE() != null ? 1 : 0), reg);
		});
	}

	@Override
	public Reg visitLiteralCharacter(LiteralCharacterContext ctx) {
		return regPool.withReg((reg) -> {
			emit(ctx.getText(), OpCode.loadI, new Num(ctx.CHAR().getText().codePointAt(1)), reg);
			emit(ctx.getText(), OpCode.i2c, reg, reg);
		});
	}

	@Override
	public Reg visitLiteralInteger(LiteralIntegerContext ctx) {
		return regPool.withReg((reg) -> {
			emit(ctx.getText(), OpCode.loadI, new Num(Integer.parseInt(ctx.LITERAL10().getText())), reg);
		});
	}

	@Override
	public Reg visitParens(ParensContext ctx) {
		return visit(ctx.expr());
	}

	@Override
	public Reg visitPrefix1(Prefix1Context ctx) {
		Reg reg = visit(ctx.singleExpr());

		switch (ctx.op.getType()) {
		case BoppiLexer.MINUS:
			emit("unary -", OpCode.rsubI, reg, ZERO, reg);
			break;

		case BoppiLexer.NOT:
			emit("not", OpCode.xorI, reg, new Num(1), reg);
			break;
		}
		return reg;
	}

	@Override
	public Reg visitProgram(ProgramContext ctx) {
		emit("initialise zero register", OpCode.loadI, ZERO, ZERO_REG);

		malloc(arp, an.function.get(ctx).getLocalDataSize() + ARBASESIZE);
		emit("construct main AR", OpCode.addI, arp, new Num(ARBASESIZE), arp);
		visitChildren(ctx);

		cleanScope(an.function.get(ctx));

		emit("deconstruct main AR", OpCode.subI, arp, new Num(ARBASESIZE), arp);
		regPool.withReg((temp) -> {
			free(temp, arp);
		});

		return null;
	}

	@Override
	public Reg visitRead(ReadContext ctx) {
		if (ctx.variable().size() > 1) {
			for (VariableContext expr : ctx.variable()) {
				visit(expr);
				if (SimpleType.BOOL.equals(an.types.get(expr)) || SimpleType.INT.equals(an.types.get(expr))) {
					regPool.withReg((temp) -> {
						emit("", OpCode.in, new Str(""), temp);
						emit("var " + expr.getText(), OpCode.storeAI, temp, getVar(expr),
								new Num(an.variables.get(expr).getOffset()));
					});
				} else if (SimpleType.CHAR.equals(an.types.get(expr))) {
					// Get input until at least 1 character
					Label getTarget = makeLabel("lcin_l");
					Label continueTarget = makeLabel("lcin_e");

					regPool.withReg((temp, count) -> {
						emit("", getTarget, OpCode.cin, new Str(""));
						emit("str length", OpCode.pop, count);
						emit("repeat if 0 length", OpCode.cbr, count, continueTarget, getTarget);

						// Get character
						emit("pop whole string", continueTarget, OpCode.cpop, temp);
						emit("var " + expr.getText(), OpCode.cstoreAI, temp, getVar(expr),
								new Num(an.variables.get(expr).getOffset()));

						// Pop all remaining characters
						Label loopTarget = makeLabel("lcpop_l");
						Label iterTarget = makeLabel("lcpop_c");
						Label stopTarget = makeLabel("lcpop_e");
						emit("pop remaining", loopTarget, OpCode.subI, count, new Num(1), temp);
						emit("pop remaining", OpCode.cbr, count, iterTarget, stopTarget);
						emit("pop remaining", iterTarget, OpCode.cpop, temp);
						emit("pop remaining", OpCode.jumpI, loopTarget);
						emit("pop remaining", stopTarget, OpCode.nop);
					});
				} else {
					emit("reading unknown type", OpCode.haltI, new Num(1));
				}
			}

			return null;
		} else {
			VariableContext expr = ctx.variable(0);
			visit(expr);
			if (SimpleType.BOOL.equals(an.types.get(expr)) || SimpleType.INT.equals(an.types.get(expr))) {
				return regPool.withReg((result) -> {
					emit("", OpCode.in, new Str(""), result);
					emit("var " + expr.getText(), OpCode.storeAI, result, getVar(expr),
							new Num(an.variables.get(expr).getOffset()));
				});
			} else if (SimpleType.CHAR.equals(an.types.get(expr))) {
				return regPool.withReg((count, result) -> {
					// Get input until at least 1 character
					Label getTarget = makeLabel("lcin_l");
					Label continueTarget = makeLabel("lcin_e");

					emit("", getTarget, OpCode.cin, new Str(""));
					emit("str length", OpCode.pop, count);
					emit("repeat if 0 length", OpCode.cbr, count, continueTarget, getTarget);

					// Get character
					emit("pop whole string", continueTarget, OpCode.cpop, result);
					emit("var " + expr.getText(), OpCode.cstoreAI, result, getVar(expr),
							new Num(an.variables.get(expr).getOffset()));

					// Pop all remaining characters
					Label loopTarget = makeLabel("lcpop_l");
					Label iterTarget = makeLabel("lcpop_c");
					Label stopTarget = makeLabel("lcpop_e");
					regPool.withReg((temp) -> {
						emit("pop remaining", loopTarget, OpCode.subI, count, new Num(1), count);
						emit("pop remaining", OpCode.cbr, count, iterTarget, stopTarget);
						emit("pop remaining", iterTarget, OpCode.cpop, temp);
						emit("pop remaining", OpCode.jumpI, loopTarget);
						emit("pop remaining", stopTarget, OpCode.nop);
					});
				});
			} else {
				emit("reading unknown type", OpCode.haltI, new Num(1));
				return null;
			}
		}
	}

	@Override
	public Reg visitVariableExpr(VariableExprContext ctx) {
		Type type = an.types.get(ctx);
		Num offset = new Num(an.variables.get(ctx.variable()).getOffset());

		return regPool.withReg((result) -> {
			if (SimpleType.CHAR.equals(type))
				emit("get " + ctx.variable().getText(), OpCode.cloadAI, getVar(ctx.variable()), offset, result);
			else
				emit("get " + ctx.variable().getText(), OpCode.loadAI, getVar(ctx.variable()), offset, result);

			if (type instanceof ReferenceType) {
				regPool.withReg((r1, r2) -> {
					String comment = "add new ref for " + ctx.variable().getText();

					incRef(r2, result);

					if (type instanceof FunctionType) {
						emit(comment, OpCode.loadAI, result, OFFSET_FUNCREF_ARP, r1);

						incARReferences(r2, r1);
					}
				});
			}

		});
	}

	@Override
	public Reg visitWhile(WhileContext ctx) {
		Label toLoop = makeLabel("while_t");
		Label toCond = makeLabel("while_f");
		Label toEnd = makeLabel("while_e");

		emit("to condition", OpCode.jumpI, toCond);
		emit("loop target", toLoop, OpCode.nop);
		visit(ctx.onTrue);

		emit("condition target", toCond, OpCode.nop);
		emit("", OpCode.cbr, visit(ctx.cond), toLoop, toEnd);
		emit("end target", toEnd, OpCode.nop);

		return null;
	}

	@Override
	public Reg visitWrite(WriteContext ctx) {
		if (ctx.expr().size() > 1) {
			for (ExprContext expr : ctx.expr()) {
				Reg result = visit(expr);

				regPool.blockReg(result, () -> {
					Type type = an.types.get(expr);

					if (SimpleType.BOOL.equals(type) || SimpleType.INT.equals(type)) {
						emit("", OpCode.out, new Str(""), result);
					} else if (SimpleType.CHAR.equals(type)) {
						regPool.withReg((temp) -> {
							emit("push character", OpCode.cpush, result);
							emit("load 1", OpCode.loadI, new Num(1), temp);
							emit("push 1", OpCode.push, temp);
							emit("print character", OpCode.cout, new Str(""));
						});
					} else {
						emit("writing unknown type", OpCode.haltI, new Num(1));
					}

					if (!expr.equals(ctx.expr(ctx.expr().size() - 1)))
						if (type instanceof ReferenceType)
							decrementReference(type, result);
				});
			}
			return null;
		} else {
			ExprContext expr = ctx.expr(0);
			Reg result = visit(expr);

			return regPool.blockReg(result, () -> {
				Type type = an.types.get(expr);

				if (SimpleType.BOOL.equals(type) || SimpleType.INT.equals(type)) {
					emit("", OpCode.out, new Str(""), result);
					return result;
				} else if (SimpleType.CHAR.equals(type)) {
					regPool.withReg(result, (temp) -> {
						emit("push character", OpCode.cpush, result);
						emit("load 1", OpCode.loadI, new Num(1), temp);
						emit("push 1", OpCode.push, temp);
						emit("print character", OpCode.cout, new Str(""));
					});
					return result;
				} else {
					emit("writing unknown type", OpCode.haltI, new Num(1));
					return null;
				}
			});
		}
	}
}
